from enum import Enum
from typing import Literal, Union

from pydantic import BaseModel, Field


class PromptTypes(Enum):
    SINGLE_GENERATION: str = "single_generation"
    CHAT_COMPLETION: str = "chat_completion"


class PromptMeta(BaseModel):
    temperature: float | None = 0.2
    tokens_limit: int | None = 8096
    stop_words: list[str] | None = None
    model: str | None = Field(default=None, examples=[None])


class PromptModel(BaseModel):
    job_id: str
    priority: int | None = Field(default=None, examples=[None])
    meta: PromptMeta
    content: str


class ChatCompletionUnit(BaseModel):
    """A model for element of chat completion"""
    role: str
    content: str


class ChatCompletionModel(BaseModel):
    """A model for chat completion order"""
    job_id: str
    priority: int | None = Field(default=None, examples=[None])
    source: str = "local"
    meta: PromptMeta
    messages: list[ChatCompletionUnit]

    @classmethod
    def from_prompt_model(cls, prompt_model: PromptModel) -> 'ChatCompletionModel':
        # Создаем первое сообщение из содержимого PromptModel
        initial_message = ChatCompletionUnit(
            role="user",  # Или другой подходящий role
            content=prompt_model.content
        )
        # Возвращаем новый экземпляр ChatCompletionModel
        return cls(
            job_id=prompt_model.job_id,
            priority=prompt_model.priority,
            meta=prompt_model.meta,
            messages=[initial_message]
        )


class PromptTransactionModel(BaseModel):
    prompt: PromptModel
    prompt_type: Literal[PromptTypes.SINGLE_GENERATION.value]


class ChatCompletionTransactionModel(BaseModel):
    prompt: ChatCompletionModel
    prompt_type: Literal[PromptTypes.CHAT_COMPLETION.value]


class PromptWrapper(BaseModel):
    prompt: Union[PromptTransactionModel, ChatCompletionTransactionModel] = Field(..., discriminator='prompt_type')


class ResponseModel(BaseModel):
    content: str


class LLMResponse(BaseModel):
    job_id: str
    text: str


class TextEmbedderRequest(BaseModel):
    job_id: str
    inputs: str
    truncate: bool


class ToEmbed(BaseModel):
    inputs: str
    truncate: bool


class TextEmbedderResponse(BaseModel):
    embeddings: list[float]



import logging
import time
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Type, Callable
from uuid import uuid4

from protollm_sdk.jobs.result_storage import ResultStorage


class InvokeType(Enum):
    """
    Enum for job invocation type
    """
    Blocking = auto()
    Worker = auto()


class JobResult(ABC):
    """
    Abstract class for job result
    """

    def __init__(self, storage: ResultStorage, job_name: str, job_id: str):
        """
        Initialize JobResult

        :param storage: storage for arbitrary job
        :type storage: ResultStorage
        :param job_name: job name
        :type job_name: str
        :param job_id: job id
        :type job_id: str
        """
        self.job_id = job_id
        self.storage = storage.for_job(job_name)

    @abstractmethod
    def get_result(self) -> dict:
        """
        Return result of job

        :return: dict
        """
        ...


class BlockingJobResult(JobResult):
    """
    Class for blocking job result.
    Executed in the caller thread, and the result is accessible immediately.
    """

    def __init__(self, storage: ResultStorage, job_name: str, job_id: str):
        """
        Initialize BlockingJobResult

        :param storage: storage for job
        :type storage: ResultStorage
        :param job_name: job name
        :type job_name: str
        :param job_id: job id
        :type job_id: str
        """
        super().__init__(storage, job_name, job_id)
        self.result = self._get_result()

    def get_result(self) -> dict:
        """
        Return result of job

        :return: dict
        """
        return self.result

    def _get_result(self) -> dict:
        t = time.monotonic()
        timeout = 5 * 60
        wait = 0.25

        while time.monotonic() - t < timeout:
            result = self._ping_result()
            if result is not None:
                return result
            time.sleep(wait)

        raise TimeoutError(f"Couldn't retrieve result in {timeout // 60}min {timeout % 60}sec")

    def _ping_result(self) -> dict | None:
        return self.storage.load_dict(self.job_id)


class WorkerJobResult(JobResult):
    """
    Class for worker job result
    Executed on celery worker. Result will be awaited.
    """

    def __init__(self, storage: ResultStorage, job_name: str, job_id: str):
        """
        Initialize WorkerJobResult

        :param storage: storage for job
        :type storage: ResultStorage
        :param job_name: job name
        :type job_name: str
        :param job_id: job id
        :type job_id: str
        """
        super().__init__(storage, job_name, job_id)

    def get_result(self) -> dict:
        """
        Return result of job

        :return: dict
        """
        t = time.monotonic()
        timeout = 10 * 60
        wait = 0.5

        while time.monotonic() - t < timeout:
            result = self._ping_result()
            if result is not None:
                return result
            time.sleep(wait)

        raise TimeoutError(f"Couldn't retrieve result in {timeout // 60}min {timeout % 60}sec")

    def _ping_result(self) -> dict | None:
        """
        Ping result of job

        :return: dict | None
        """
        return self.storage.load_dict(self.job_id)


class JobInvoker:
    """
    Class for invoking job
    """

    def __init__(
            self,
            abstract_celery_task: Callable[[type, str, ...], None] | None,
            result_storage: ResultStorage,
            invoke_type: InvokeType = InvokeType.Worker):
        """
        Initialize JobInvoker

        :param abstract_celery_task: reference to celery.task
        :type abstract_celery_task: Callable[[type, str, ...], None]
        :param result_storage: storage for job
        :type result_storage: ResultStorage
        :param invoke_type: job invocation type
        :type invoke_type: InvokeType
        """

        self._logger = logging.getLogger(self.__class__.__name__)
        self._task = abstract_celery_task
        self._result_storage = result_storage
        self._invoke_type = invoke_type

    def invoke(self, job_class: Type, **kwargs) -> JobResult:
        """
        Invoke job based on specified invoke type

        :param job_class: job class
        :type job_class: Type
        :param kwargs: additional arguments
        :type kwargs: dict
        :return: JobResult
        """

        if self._task is None:
            msg = "Calling JobInvoker without abstract task."
            self._logger.error(msg)
            raise Exception(msg)

        task_id = str(uuid4())
        job_name = job_class.__name__ if isinstance(job_class, type) else job_class.__class__.__name__

        self._logger.info(f"Invoking '{job_name}' with task id'{task_id}'.")

        match self._invoke_type:
            case InvokeType.Blocking:
                self._task(job_class, task_id, **kwargs)
                return BlockingJobResult(self._result_storage, job_name, task_id)
            case InvokeType.Worker:
                self._task.apply_async(args=(job_class, task_id), kwargs=kwargs)
                return WorkerJobResult(self._result_storage, job_name, task_id)
            case _:
                msg = f"Unknown invoke type: {self._invoke_type}."
                self._logger.error(msg)
                raise NotImplementedError(msg)



import logging

from celery import Celery
from kombu.serialization import registry

from protollm_sdk.celery.config import CeleryConfig
from protollm_sdk.celery.constants import JOB_NAME2CLASS
from protollm_sdk.jobs.job import Job
from protollm_sdk.jobs.utility import construct_job_context


def init_celery(celery_config: CeleryConfig) -> Celery:
    init_args, init_kwargs = celery_config.init_args

    celery = Celery(*init_args, **init_kwargs)

    if celery_config.conf_update:
        celery.conf.update(**celery_config.conf_update)

    if celery_config.formats:
        for f in celery_config.formats:
            registry.enable(f)

    return celery

celery_config = CeleryConfig()
celery = init_celery(celery_config)


logger = logging.getLogger(__name__)


@celery.task(**celery_config.task_kwargs)
def task_test(task_class: str, task_id: str, **kwargs): # noqa
    ctx = construct_job_context(task_class, abstract_task)
    if job_class := JOB_NAME2CLASS.get(task_class):
        job = job_class()
    else:
        msg = f"Error in task '{task_id}'. Unknown job class: '{task_class}'."
        logger.error(msg)
        raise Exception(msg)

    logger.info(f"Starting task '{task_id}'. Job '{task_class}'.")

    forecast = job.run(task_id=task_id, ctx=ctx, **kwargs)
    return forecast


@celery.task(**celery_config.task_kwargs)
def abstract_task(task_class: type[Job], task_id: str, **kwargs):
    class_name = task_class.__name__ if isinstance(task_class, type) else task_class.__class__.__name__
    ctx = construct_job_context(class_name, abstract_task)
    logger.info(f"Starting task '{task_id}'. Job '{class_name}'.")
    job_object = task_class() if isinstance(task_class, type) else task_class
    job_object.run(task_id=task_id, ctx=ctx, **kwargs)



from dataclasses import dataclass

import logging

from protollm_sdk.jobs.outer_llm_api import OuterLLMAPI
from protollm_sdk.jobs.job_invoke import JobInvoker
from protollm_sdk.jobs.vector_db import VectorDB
from protollm_sdk.jobs.result_storage import ResultStorage
from protollm_sdk.jobs.text_embedder import TextEmbedder
from protollm_sdk.jobs.llm_api import LLMAPI

logger = logging.getLogger(__name__)


@dataclass
class JobContext:
    """
    The class contains contextual services for executing Job
    """
    llm_api: LLMAPI
    outer_llm_api: OuterLLMAPI
    text_embedder: TextEmbedder
    result_storage: ResultStorage
    vector_db: VectorDB
    job_invoker: JobInvoker





import logging

from protollm_sdk.jobs.job import Job
from protollm_sdk.jobs.job_context import JobContext
from protollm_sdk.jobs.job_invoke import InvokeType
from protollm_sdk.models.job_context_models import TextEmbedderRequest, PromptModel


class TextEmbedderJob(Job):
    """Defining the embedder class inherited from Job"""
    def run(self, task_id: str, ctx: JobContext, **kwargs):
        """Redefining the run function"""
        request = TextEmbedderRequest(**kwargs)
        resp = ctx.text_embedder.inference(request)
        return resp


class ResultStorageJob(Job):
    """Defining the storage class inherited from Job"""

    def run(self, task_id: str, ctx: JobContext, **kwargs):
        """Redefining the run function"""
        ctx.result_storage.save_dict(job_id=task_id, result=kwargs)


class LLMAPIJob(Job):
    """Defining the llm class inherited from Job"""

    def run(self, task_id: str, ctx: JobContext, **kwargs):
        """Redefining the run function"""
        request = PromptModel(**kwargs)
        resp = ctx.llm_api.inference(request)
        return resp


class OuterLLMAPIJob(Job):
    """Defining the llm class inherited from Job"""
    def run(self, task_id: str, ctx: JobContext, **kwargs):
        """Redefining the run function"""
        request = PromptModel(**kwargs)
        resp = ctx.outer_llm_api.inference(request)
        ctx.result_storage.save_dict(task_id, resp)
        # TODO add return??


class VectorDBJob(Job):
    """Defining the vector_db class inherited from Job"""

    def run(self, task_id: str, ctx: JobContext, **kwargs):
        """Redefining the run function"""
        resp = ctx.vector_db.api_v1()
        return resp

# class LangchainLLMAPIJob(Job):
#     """Defining the langchain class inherited from Job"""
#      def run(self, task_id: str, ctx: JobContext, **kwargs):
#         """Redefining the run function"""
#         request = LLMRequest(**kwargs)
#         request.job_id = task_id
#         resp = ctx.langchain_llm_api._call(request)
#         return resp


class TestInvocationJob(Job):
    def run(self, task_id: str, ctx: JobContext, **kwargs):
        logger = logging.getLogger(self.__class__.__name__)
        logger.warning(f"Running {self.__class__.__name__}")

        my_kwargs = dict(my_kwarg="simple_value")

        # This lets you manipulate the behaviour of the job invoker.
        # Normally, you shouldn't set this manually, but you can for the testing purposes.
        # - InvokeType.Worker (default on server) - the called job is executed on the remote worker.
        #   Call returns immediately, so you can do something before waiting for the result.
        # - InvokeType.Blocking - The called job is executed on the caller therad, and the result is accessible right afther the call returns.
        ctx.job_invoker._invoke_type = (
            InvokeType.Worker
            if kwargs.get("invoker", "blocking") == "worker"
            else InvokeType.Blocking
        )

        # Job calling is executed via passing the job class reference and the corresponding kwargs.
        job_result = ctx.job_invoker.invoke(ResultStorageJob, **my_kwargs)
        logger.warning(f"{self.__class__.__name__} sent {ResultStorageJob.__class__.__name__} task with kwargs {my_kwargs}")

        # The get_result behaviour depends on the InvokeType (described above).
        result = job_result.get_result()
        logger.warning(f"{self.__class__.__name__} got result {result}")

        ctx.result_storage.save_dict(task_id, result)



from protollm_sdk.config import Config
from protollm_sdk.jobs.result_storage import ResultStorage
from protollm_sdk.object_interface import RedisWrapper


def get_reddis_wrapper():
    """
    Create RedisWrapper object for working with Redis

    :return: RedisWrapper object
    :rtype: RedisWrapper
    """
    return RedisWrapper(
        redis_host=Config.redis_host,
        redis_port=Config.redis_port
    )


def load_result(rd: RedisWrapper, job_id: str, prefix: str or None) -> bytes:
    """
    Load result from Redis by job_id and prefix (job_name).
    The code returns bytes. If you want to convert these bytes into a ResponseModel,
    first decode the bytes into a string using `.decode()`, then create the model
    like this:
    model_text = byte_data.decode()
    response_model = ResponseModel.model_validate_json(model_text)

    :param rd: RedisWrapper object
    :type rd: RedisWrapper
    :param job_id: uuid of the job
    :type job_id: str
    :param prefix: prefix of type of job or jon_name
    :type prefix: str or None
    :return: result
    :rtype: bytes
    """
    resp = rd.get_item(ResultStorage.build_key(job_id, prefix))
    return resp



from typing import Any

from protollm_sdk.config import Config
from protollm_sdk.utils.singleton import Singleton


__all__ = ["CeleryConfig"]

_ARG_UNSET = object()


class _Defaults:
    conf_update = dict(
        task_serializer="pickle",
        accept_content=["pickle"],
        result_serializer="pickle",
    )

    formats = ["json", "application/json"]


class CeleryConfig(metaclass=Singleton):
    def __init__(
        self,
        queue: str | None = None,
        mq_usr: str | None = None,
        mq_pwd: str | None = None,
        mq_host: str | None = None,
        mq_port: int | None | object = _ARG_UNSET,
        redis_host: str | None = None,
        redis_port: int | None | object = _ARG_UNSET,
        celery_init_kwargs: dict | None = None,
        conf_update: dict[str, Any] | None | object = _ARG_UNSET,
        formats: list[str] | None | object = _ARG_UNSET,
        task_kwargs: dict[str, Any] | None | object = _ARG_UNSET
    ):
        self.queue = queue if queue is not None else Config.celery_queue_name

        self.mq_usr = mq_usr if mq_usr is not None else Config.rabbit_mq_login
        self.mq_pwd = mq_pwd if mq_pwd is not None else Config.rabbit_mq_password
        self.mq_host = mq_host if mq_host is not None else Config.rabbit_mq_host
        self.mq_port = mq_port if mq_port is not _ARG_UNSET else Config.rabbit_mq_port

        self.redis_host = redis_host if redis_host is not None else Config.redis_host
        self.redis_port = (
            redis_port
            if redis_port is not _ARG_UNSET
            else Config.redis_port
        )

        self.celery_init_kwargs = celery_init_kwargs

        self.conf_update = conf_update if conf_update is not _ARG_UNSET else _Defaults.conf_update
        self.formats = formats if formats is not _ARG_UNSET else _Defaults.formats

        self.task_kwargs = (task_kwargs or {}) if task_kwargs is not _ARG_UNSET else dict(
            queue = self.queue,
        )

    @property
    def init_args(self) -> tuple[tuple[str], dict[str, str]]:
        static_kwargs = dict(broker=self.mq_url, backend=self.redis_url)

        return (
            (self.queue, ),
            static_kwargs if self.celery_init_kwargs is None else dict(static_kwargs, **self.celery_init_kwargs),
        )

    @property
    def mq_url(self) -> str:
        schema = "amqp"
        creds = f"{self.mq_usr}:{self.mq_pwd}"
        base = f"{self.mq_host}:{self.mq_port}" if self.mq_port is not None else self.mq_host

        return f"{schema}://{creds}@{base}/"

    @property
    def redis_url(self) -> str:
        schema = "redis"
        base = f"{self.redis_host}:{self.redis_port}" if self.redis_port is not None else self.redis_host

        return f"{schema}://{base}"



import json
import logging
from contextlib import asynccontextmanager, contextmanager

import aioredis
import redis

logger = logging.getLogger(__name__)


class RedisWrapper:
    def __init__(self, redis_host: str, redis_port: str | int | None = None):
        self.url = f"redis://{redis_host}:{redis_port}" if redis_port else redis_host

    @asynccontextmanager
    async def get_redis_async(self):
        rd = aioredis.from_url(self.url)
        try:
            yield rd
        finally:
            await rd.close()

    @contextmanager
    def _get_redis(self):
        rd = redis.Redis.from_url(self.url)
        try:
            yield rd
        finally:
            rd.close()

    def save_item(self, key, item: dict) -> None:
        try:
            with self._get_redis() as redis:
                redis.set(key, json.dumps(item))
                redis.publish(key, 'set')
                logger.info(f"The result with the prefix {key} has been successfully "
                            f"written to the redis")
        except Exception as ex:
            msg = f"Saving the result with the {key} prefix has been interrupted. Error: {ex}."
            logger.info(msg)
            raise Exception(msg)

    def get_item(self, key) -> bytes:
        try:
            with self._get_redis() as redis:
                response = redis.get(key)
                logger.info(f"The item with the {key} prefix was successfully retrieved from the redis")
                return response
        except Exception as ex:
            msg = f"The receipt of the element with the {key} prefix was interrupted. Error: {ex}"
            logger.info(msg)
            raise Exception(msg)

    def check_key(self, key) -> bool:
        try:
            with self._get_redis() as redis:
                response = redis.get(key)
                logger.info(f"The request for the {key} key has been successfully processed.")
                return True if response is not None else False
        except Exception as ex:
            msg = f"An error occurred while processing the {key} key. Error: {ex}"
            logger.info(msg)
            raise Exception(msg)

    async def wait_item(self, key, timeout: float = 60) -> bytes:
        """
            Waits for an item to appear in the specified Redis channel.

            Note:
                There is a very small chance that the message might be published
                before the intended channel is subscribed to. This could result
                in the `wait_item` method getting stuck in a loop.
            """
        try:
            async with self.get_redis_async() as redis:
                pubsub = redis.pubsub()
                await pubsub.subscribe(f"{key}")
                while True:
                    message = await pubsub.get_message(ignore_subscribe_messages=True, timeout=timeout)
                    if message is not None:
                        event = message['data']
                        if event == b'set':
                            value = await redis.get(key)
                            if value is not None:
                                logger.info(f"The item with the {key} prefix was successfully retrieved from the redis")
                                return value
        except Exception as ex:
            msg = f"The receipt of the element with the {key} prefix was interrupted. Error: {ex}"
            logger.info(msg)
            raise Exception(msg)



import json
import logging
from contextlib import contextmanager
import pika
from sqlalchemy import Boolean

logger = logging.getLogger(__name__)


class RabbitMQWrapper:
    def __init__(self, rabbit_host: str, rabbit_port: int, rabbit_user: str, rabbit_password: str, virtual_host: str = '/'):
        """
        Initialize RabbitMQ wrapper.

        :param rabbit_host: RabbitMQ host
        :param rabbit_port: RabbitMQ port
        :param rabbit_user: RabbitMQ username
        :param rabbit_password: RabbitMQ password
        :param virtual_host: RabbitMQ virtual host
        """
        self.connection_params = pika.ConnectionParameters(
            host=rabbit_host,
            port=rabbit_port,
            virtual_host=virtual_host,
            credentials=pika.PlainCredentials(rabbit_user, rabbit_password)
        )

    @contextmanager
    def get_channel(self):
        """
        Provide a channel for RabbitMQ operations.

        :yield: pika channel
        """
        connection = pika.BlockingConnection(self.connection_params)
        channel = connection.channel()
        try:
            yield channel
        finally:
            channel.close()
            connection.close()

    def publish_message(self, queue_name: str, message: dict, durable: bool=True):
        """
        Publish a message to a specified queue with an optional priority.

        :param queue_name: Name of the queue to publish to
        :param message: Message to publish (dictionary will be serialized to JSON)
        :param priority: Optional priority of the message (0-255)
        :param durable: Optional type of durability of queue
        """
        try:
            with self.get_channel() as channel:
                arguments = {'x-max-priority': 5}
                priority = message.get("kwargs").get("prompt").get("priority")

                channel.queue_declare(queue=queue_name, durable=durable, arguments=arguments)

                properties = pika.BasicProperties(
                    delivery_mode=2,
                    priority=priority
                )
                channel.basic_publish(
                    exchange='',
                    routing_key=queue_name,
                    body=json.dumps(message),
                    properties=properties
                )
                logger.info(
                    f"Message published to queue '{queue_name}' with priority {priority if priority is not None else 'None'}")
        except Exception as ex:
            logger.error(f"Failed to publish message to queue '{queue_name}'. Error: {ex}")
            raise Exception(f"Failed to publish message to queue '{queue_name}'. Error: {ex}") from ex

    def consume_messages(self, queue_name: str, callback, durable: bool=True):
        """
        Start consuming messages from a specified queue.

        :param queue_name: Name of the queue to consume from
        :param callback: Callback function to process messages
        :param durable: Optional type of durability of queue
        """
        try:
            connection = pika.BlockingConnection(self.connection_params)
            channel = connection.channel()

            channel.queue_declare(queue=queue_name, durable=durable, arguments={"x-max-priority": 5})

            channel.basic_consume(
                queue=queue_name,
                on_message_callback=callback,
                auto_ack=True
            )
            logger.info(f"Started consuming messages from queue '{queue_name}'")
            channel.start_consuming()
        except Exception as ex:
            logger.error(f"Failed to consume messages from queue '{queue_name}'. Error: {ex}")
            raise Exception(f"Failed to consume messages from queue '{queue_name}'. Error: {ex}")


import json
import logging
from contextlib import contextmanager

import redis

logger = logging.getLogger(__name__)


class ResultStorage:
    """
    Class provides an object interface to a temporary (cache) fast (in memory)
    key-value storage of results - Redis.
    It writes the results to the storage by key, which consists of the name of the agent job_name and
    the identifier of the current task (job_id)
    """

    def __init__(self, redis_host: str, redis_port: str | int | None = None, prefix: str | None = None):
        """
        Initialize ResultStorage

        :param redis_host: host of the redis
        :type redis_host: str
        :param redis_port: port of the redis
        :type redis_port: str | int | None
        :param prefix: prefix for the key
        :type prefix: str | None
        """
        self.url = f"redis://{redis_host}:{redis_port}" if redis_port is not None else f"redis://{redis_host}"
        self.prefix = prefix or ""

    def for_job(self, job_name: str) -> "ResultStorage":
        """
        Clones the connection to redis, but for another job
        (useful if you call one job from another, and the key prefix in the storage should be different)

        :param job_name: job name
        :type job_name: str
        :return: ResultStorage
        """
        return ResultStorage(redis_host=self.url, redis_port=None, prefix=job_name)

    @contextmanager
    def _get_redis(self):
        """
        Get redis connection
        """
        rd = redis.Redis.from_url(self.url)
        try:
            yield rd
        finally:
            rd.close()

    def _get_key(self, job_id: str) -> str:
        """
        Format the string to get the key for the agent
        for which the storage instance was created, and for the current job_id

        :param job_id: job id
        :type job_id: str
        :return: Formatted key
        """
        return self.build_key(job_id, self.prefix)

    def save_dict(self, job_id: str, result: dict):
        """
        Save the dictionary to the storage by key

        :param job_id: job id
        :type job_id: str
        :param result: true dictionary, not a dictionary serialized to a byte string
        :type result: dict
        :return: None
        """
        key = self._get_key(job_id)
        try:
            with self._get_redis() as rs:
                rs.set(key, json.dumps(result))
                logger.info(f"The result with the key {key} has been successfully written to the redis")
        except Exception as ex:
            msg = f"Saving the result with the key {key} has been interrupted. Error: {ex}."
            logger.info(msg)
            raise Exception(msg) from ex

    def load_dict(self, job_id: str) -> dict:
        """
        Load the dictionary from the storage by key

        :param job_id: job id
        :type job_id: str
        :return: true dictionary, not a dictionary serialized to a byte string
        """
        key = self._get_key(job_id)
        try:
            with self._get_redis() as rs:
                result = rs.get(key)
                logger.info(f"Data by key {key} has been successfully read from redis.")
                if isinstance(result, (bytes, str)):
                    return json.loads(result)
                return result
        except Exception as ex:
            msg = f"Failed to load the result with the key {key} from redis. Error: {ex}."
            logger.error(msg)
            raise Exception(msg) from ex

    @staticmethod
    def build_key(job_id: str, prefix: str or None) -> str:
        """
        Format any suffix and prefix according to the key rules

        :param job_id: job id
        :type job_id: str
        :param prefix: prefix
        :type prefix: str or None
        :return: str
        """
        return job_id if prefix is None else f"{prefix}:{job_id}"



from urllib.parse import urljoin

import httpx


class VectorDB:
    """
    VectorDB client
    """

    def __init__(self, vector_bd_host: str, vector_db_port: str | int | None = None):
        """
        Initialize VectorDB

        :param vector_bd_host: host of vector db
        :type vector_bd_host: str
        :param vector_db_port: port of vector db
        :type vector_db_port: str | int | None
        """
        self.url = f"http://{vector_bd_host}:{vector_db_port}" if vector_db_port is not None else f"http://{vector_bd_host}"
        self.client = httpx.Client()

    def api_v1(self):
        """
        Get api v1

        :return: response
        """
        response = self.client.get(
            urljoin(self.url, "/api/v1"),
            headers={"Content-type": "application/json"},
            timeout=10 * 60
        )
        return response.json()



import logging
from urllib.parse import urljoin

import httpx

from protollm_sdk.models.job_context_models import TextEmbedderRequest, TextEmbedderResponse, ToEmbed

logger = logging.getLogger(__name__)


class TextEmbedder:
    """
    Class provides an object interface to a text embedder.
    """

    def __init__(
            self,
            text_emb_host: str,
            text_emb_port: str | int | None = None,
            timeout_sec: int = 10 * 60,
    ):
        """
        Initialize TextEmbedder

        :param text_emb_host: host of the text embedder
        :type text_emb_host: str
        :param text_emb_port: port of the text embedder
        :type text_emb_port: str | int | None
        :param timeout_sec: timeout in seconds
        :type timeout_sec: int
        """
        self.path = f"http://{text_emb_host}:{text_emb_port}" if text_emb_port is not None else f"http://{text_emb_host}"
        self.timeout_sec = timeout_sec
        self.client = httpx.Client()

    def inference(self, request: TextEmbedderRequest) -> TextEmbedderResponse:
        """
        Create an embedding for the input text

        :param request: request
        :type request: TextEmbedderRequest
        :return: TextEmbedderResponse
        """
        try:
            emb = ToEmbed(inputs=request.inputs, truncate=request.truncate)
            response = self.client.post(
                urljoin(self.path, "/embed"),
                headers={"Content-type": "application/json"},
                json=emb.model_dump(),
                timeout=self.timeout_sec
            )
            if response.status_code == 500:
                raise ConnectionError('The LLM server is not available.')
            elif response.status_code == 422:
                raise ValueError(f'Data model validation error. {response.json()}')
            result = response.json()[0]
            logger.info("The embedding has been completed successfully.")
            text_embedder_result = {"embeddings": result}
            return TextEmbedderResponse(**text_embedder_result)
        except Exception as ex:
            msg = f"The embedding was interrupted. Error: {ex}."
            logger.info(msg)
            raise Exception(msg)



from protollm_sdk.celery.job import TextEmbedderJob, LLMAPIJob, ResultStorageJob, VectorDBJob, OuterLLMAPIJob

JOBS = {TextEmbedderJob, LLMAPIJob, ResultStorageJob, VectorDBJob, OuterLLMAPIJob}

JOB_NAME2CLASS = {cls.__name__: cls for cls in JOBS}



from abc import ABC, abstractmethod
from typing import TypeVar

from protollm_sdk.jobs.job_context import JobContext


class Job(ABC):
    """
    Job interface for integration with outer modules to SDK.
    All the required data for job should be passed and parameters should be defined in advance,
    this also applies to the run method
    """

    @abstractmethod
    def run(self, job_id: str, ctx: JobContext, **kwargs):
        """
        Run the job. The job can use a number of functions defined in the module and service functions from ctx.
        After that, using the ctx, it saves the result to Redis.
        The method should not return any data. If an error occurs,
        it is thrown inside the run method via the `raise`, `raise ex` or `raise … from ex` statement.

        :param job_id: job id
        :type job_id: str
        :param ctx: contextual services
        :type ctx: JobContext
        :param kwargs: The data and parameters required to complete the task are set via key arguments
        :return: None
        :raises TypeError: Error

        """
        pass


TJob = TypeVar('TJob', bound=Job)



import os
from dataclasses import dataclass


@dataclass(frozen=True)
class Config:
    llm_api_host = os.environ.get("LLM_API_HOST", "localhost")
    llm_api_port = os.environ.get("LLM_API_PORT", "6672")

    outer_llm_key = os.environ.get("OUTER_LLM_KEY", "sk-or-vv-c49f40fdb086053ec32c6ae2723b8d222cb7767f3b98527e7ae282986e7d33ed")

    redis_host = os.environ.get("REDIS_HOST", "localhost")
    redis_port = os.environ.get("REDIS_PORT", "6379")

    rabbit_mq_host = os.environ.get("RABBIT_HOST", "localhost")
    rabbit_mq_port = os.environ.get("RABBIT_PORT", "5672")
    rabbit_mq_login = os.environ.get("RABBIT_MQ_LOGIN", "admin")
    rabbit_mq_password = os.environ.get("RABBIT_MQ_PASSWORD", "admin")

    text_embedder_host = os.environ.get("TEXT_EMB_HOST", "localhost")
    text_embedder_port = os.environ.get("TEXT_EMB_PORT", "9942")

    vector_bd_host = os.environ.get("VECTOR_HOST", "localhost")
    vector_db_port = os.environ.get("VECTOR_PORT", "9941")

    job_invocation_type = os.environ.get("JOB_INVOCATION_TYPE", "worker")

    celery_queue_name = os.environ.get("CELERY_QUEUE_NAME", "celery")

    @classmethod
    def reload_invocation_type(cls):
        cls.job_invocation_type = os.environ.get("JOB_INVOCATION_TYPE", "worker")



import json
import logging
from urllib.parse import urljoin

import httpx

from protollm_sdk.models.job_context_models import PromptModel, ResponseModel, ChatCompletionModel

logger = logging.getLogger(__name__)


class LLMAPI:
    """
    Class for using large language models deployed on the cluster
    """

    def __init__(
            self,
            llm_api_host: str,
            llm_api_port: str | int | None = None,
            timeout_sec: int = 10 * 60,
    ):
        """
        Initialize LLMAPI

        :param llm_api_host: Host of the LLM API
        :type llm_api_host: str
        :param llm_api_port: Port of the LLM API
        :type llm_api_port: str | int | None
        :param timeout_sec: Timeout in seconds
        :type timeout_sec: int
        """
        self.path = f"http://{llm_api_host}:{llm_api_port}" if llm_api_port is not None else f"http://{llm_api_host}"
        self.timeout_sec = timeout_sec
        self.client = httpx.Client()

    def inference(self, request: PromptModel, queue_name: str = None) -> ResponseModel:
        """
        Simple response to a request in the response-question format

        :param request: simple request
        :type request: PromptModel
        :param queue_name: name of the queue to which the request will be sent
        :type queue_name: str
        :return: ResponseModel
        """
        params = {}
        if queue_name is not None:
            params["queue_name"] = queue_name
        try:
            response = self.client.post(
                urljoin(self.path, "/generate"),
                headers={"Content-type": "application/json"},
                params=params,
                data=request.model_dump_json(),
                timeout=self.timeout_sec
            )
            if response.status_code == 500:
                raise ConnectionError('The LLM server is not available.')
            elif response.status_code == 422:
                raise ValueError(f'Data model validation error. {response.json()}')
            result = ResponseModel.model_validate(response.json())
            logger.info("The request has been successfully processed.")
            return result
        except Exception as ex:
            msg = f"The response generation has been interrupted. Error: {ex}."
            logger.info(msg)
            raise Exception(msg)

    def chat_completion(self, request: ChatCompletionModel, queue_name: str = None) -> ResponseModel:
        """
        Response to a request taking into account the chat history
        (the initial prompt, user requests and model responses will be sent)

        :param request: request for chat completion
        :type request: ChatCompletionModel
        :param queue_name: name of the queue to which the request will be sent
        :type queue_name: str
        :return: ResponseModel
        """
        params = {}
        if queue_name is not None:
            params["queue_name"] = queue_name
        try:
            response = self.client.post(
                urljoin(self.path, "/chat_completion"),
                headers={"Content-type": "application/json"},
                params=params,
                data=request.model_dump_json(),
                timeout=self.timeout_sec
            )
            if response.status_code == 500:
                raise ConnectionError('The LLM server is not available.')
            elif response.status_code == 422:
                raise ValueError(f'Data model validation error. {response.json()}')
            result = ResponseModel.model_validate(response.json())
            logger.info("The request has been successfully processed.")
            return result
        except Exception as ex:
            msg = f"The response generation has been interrupted. Error: {ex}."
            logger.info(msg)
            raise Exception(msg)

    def __del__(self):
        """
        Close client

        :return: None
        """
        self.client.close()



import logging
from time import sleep

from openai import OpenAI

from protollm_sdk.models.job_context_models import PromptModel, ChatCompletionModel, ResponseModel

logger = logging.getLogger(__name__)


class OuterLLMAPI:
    """
    Class for using large language models deployed on the vsegpt.ru
    """

    def __init__(
            self,
            key: str,
            timeout_sec: int = 10 * 60,
    ):
        """
        Initialize OuterLLMAPI

        :param timeout_sec: Timeout in seconds
        :type timeout_sec: int
        """
        self.model = "openai/gpt-4o-2024-08-06"
        self.key = key
        self.timeout_sec = timeout_sec
        self.client = OpenAI(
            api_key=self.key,
            base_url="https://api.vsegpt.ru/v1",
        )

    def _chat_completion(
            self,
            messages: list[dict[str, str]],
            temperature: float,
            tokens_limit: int,
            model: str or None
    ) -> str:
        """
        Set up a chat completion

        :param messages: history of messages
        :type messages: list[dict[str, str]]
        :param temperature: degree of creativity in the responses
        :type temperature: float
        :param tokens_limit: maximum number of tokens in the response
        :type tokens_limit: int
        :param model: model name
        :type model: str or None
        :return: str
        """
        sleep(1)  # TODO remake in singleton
        # GPT chats don't respond more than 1 message at second
        model = model or self.model
        response = self.client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            n=1,
            max_tokens=tokens_limit,
            timeout=self.timeout_sec)

        result = response.choices[0].message.content
        logger.info(result)
        return result

    def inference(self, request: PromptModel) -> ResponseModel:
        """
        Simple response to a request in the response-question format

        :param request: simple request
        :type request: PromptModel
        :return: ResponseModel
        """
        messages = [{"role": "user", "content": request.content}]
        try:
            result = self._chat_completion(
                messages, request.meta.temperature,
                request.meta.tokens_limit,
                model=request.meta.model
            )
            return ResponseModel(content=result)
        except Exception as ex:
            msg = f"The response generation has been interrupted. Error: {ex}."
            logger.error(msg)
            raise Exception(msg)

    def chat_completion(self, request: ChatCompletionModel) -> ResponseModel:
        """
        Response to a request taking into account the chat history
        (the initial prompt, user requests and model responses will be sent)

        :param request: request for chat completion
        :type request: ChatCompletionModel
        :return: ResponseModel
        """
        messages = [{"role": unit.role, "content": unit.content} for unit in request.messages]
        try:
            result = self._chat_completion(
                messages,
                request.meta.temperature,
                request.meta.tokens_limit,
                model=request.meta.model
            )
            return ResponseModel(content=result)
        except Exception as ex:
            msg = f"The response generation has been interrupted. Error: {ex}."
            logger.error(msg)
            raise Exception(msg)



from protollm_sdk.config import Config
from protollm_sdk.jobs.job_context import JobContext
from protollm_sdk.jobs.job_invoke import JobInvoker, InvokeType
from protollm_sdk.jobs.llm_api import LLMAPI
from protollm_sdk.jobs.outer_llm_api import OuterLLMAPI
from protollm_sdk.jobs.result_storage import ResultStorage
from protollm_sdk.jobs.text_embedder import TextEmbedder
from protollm_sdk.jobs.vector_db import VectorDB


def construct_job_context(job_name: str, abstract_task=None) -> JobContext:
    """
    Create JobContext object with object access to functions and services, based on environment variable values.

    :param job_name: job name
    :type job_name: str
    :param abstract_task: optional reference to celery.task for recursive Job calling from Job
    :return: JobContext object
    """

    llm_api = LLMAPI(Config.llm_api_host, Config.llm_api_port)

    out_llm_api = OuterLLMAPI(Config.outer_llm_key)

    text_embedder = TextEmbedder(Config.text_embedder_host, Config.text_embedder_port)

    result_storage = ResultStorage(
        redis_host=Config.redis_host,
        redis_port=Config.redis_port,
        prefix=job_name
    )

    invoke_type_str = Config.job_invocation_type.lower()
    match invoke_type_str:
        case "worker":
            invoke_type = InvokeType.Worker
        case "blocking":
            invoke_type = InvokeType.Blocking
        case _:
            raise ValueError(f"Found unknown invocation type '{invoke_type_str}'.")
    job_invoker = JobInvoker(abstract_task, result_storage, invoke_type)

    vector_db = VectorDB(vector_bd_host=Config.vector_bd_host, vector_db_port=Config.vector_db_port)

    return JobContext(
        llm_api=llm_api,
        outer_llm_api=out_llm_api,
        text_embedder=text_embedder,
        result_storage=result_storage,
        vector_db=vector_db,
        job_invoker=job_invoker,
    )
